from __future__ import annotations

from typing import *  # type: ignore

import rio


# <additional-imports>
from ..data_models import TodoItem
# </additional-imports>


# <component>
ICON_BUTTON_SIZE = 2.5


class TodoItemComponent(rio.Component):
    """
    Displays a single `TodoItem`.

    Includes the `TodoItem`'s title and creation date, a button that marks it as
    completed, and a button that deletes it.
    """

    todo_item: TodoItem
    on_completed: rio.EventHandler[[]] = None
    on_deleted: rio.EventHandler[[]] = None

    async def _mark_as_completed(self) -> None:
        # If it's already completed, there's nothing to do
        if self.todo_item.completed:
            return

        self.todo_item.completed = True
        await self.call_event_handler(self.on_completed)

        # Rio doesn't know that we modified the TodoItem, so it won't
        # automatically rebuild this component. We have to manually trigger a
        # rebuild.
        await self.force_refresh()

    def build(self) -> rio.Component:
        return rio.Card(
            rio.Row(
                # The "mark as complete" button
                rio.IconButton(
                    "material/check",
                    size=ICON_BUTTON_SIZE,
                    color="success" if self.todo_item.completed else "neutral",
                    on_press=self._mark_as_completed,
                ),
                # The title and creation date
                rio.Column(
                    rio.Text(self.todo_item.title, justify="left"),
                    rio.Text(
                        f"{self.todo_item.creation_time:%A, %x}",
                        justify="left",
                        style="dim",
                    ),
                ),
                rio.Spacer(),
                # The "delete" button
                rio.IconButton(
                    "material/delete",
                    size=ICON_BUTTON_SIZE,
                    on_press=self.on_deleted,
                ),
                spacing=0.5,
            ),
            color="background",
            elevate_on_hover=True,
            corner_radius=999,
        )


# </component>



import asyncio
import functools
import inspect
from typing import TypeVar

from uniserde import Jsonable

import rio.global_state
from rio.app_server import AppServer
from rio.components.root_components import (
    HighLevelRootComponent,
)


__all__ = ["enable_component_instantiation"]


T = TypeVar("T")
C = TypeVar("C", bound=rio.Component)


def enable_component_instantiation(func):
    @functools.wraps(func)
    def wrapper(*args, **kwargs):
        def build():
            result = func(*args, **kwargs)
            assert not inspect.isawaitable(
                result
            ), "The test function must not be async"

            return rio.Text("")

        app = rio.App(build=build)
        app_server = AppServer(
            app_=app,
            debug_mode=False,
            running_in_window=False,
            validator_factory=None,
            internal_on_app_start=None,
        )
        session = rio.Session(
            app_server,
            "<a fake session token>",
            "<a fake client ip>",
            12345,
            "<a fake user agent>",
        )
        session._decimal_separator = "."
        session._thousands_separator = ","
        session._send_message = _fake_send_message
        session._receive_message = _fake_receive_message

        rio.global_state.currently_building_session = session
        session._root_component = HighLevelRootComponent(
            build, lambda: rio.Text("")
        )

        rio.global_state.currently_building_component = session._root_component
        try:
            session._root_component.build()
        finally:
            rio.global_state.currently_building_component = None
            rio.global_state.currently_building_session = None

    return wrapper


async def _fake_send_message(message: Jsonable) -> None:
    pass


async def _fake_receive_message() -> Jsonable:
    while True:
        await asyncio.sleep(100000)



from __future__ import annotations

import os
import sys
import threading
import webbrowser
from collections.abc import Callable, Iterable
from datetime import timedelta
from pathlib import Path
from typing import *  # type: ignore

import __main__
import fastapi
import uvicorn

import rio

from . import app_server, assets, debug, maybes, utils
from .utils import ImageLike


__all__ = [
    "App",
]


def make_default_connection_lost_component() -> rio.Component:
    class DefaultConnectionLostComponent(rio.Component):
        def build(self) -> rio.Component:
            return rio.Rectangle(
                content=rio.Rectangle(
                    content=rio.Row(
                        rio.Icon(
                            "material/error",
                            fill="danger",
                            width=1.6,
                            height=1.6,
                        ),
                        rio.Text(
                            "Disconnected",
                            style=rio.TextStyle(
                                fill=self.session.theme.hud_palette.foreground,
                                font_weight="bold",
                            ),
                        ),
                        spacing=0.5,
                        margin_x=2.5,
                        margin_y=1.5,
                    ),
                    fill=self.session.theme.hud_palette.background,
                    corner_radius=99999,
                    margin_bottom=0.15,
                ),
                fill=self.session.theme.danger_palette.background,
                corner_radius=99999,
                shadow_color=self.session.theme.shadow_color,
                shadow_radius=0.6,
                shadow_offset_y=0.15,
                margin_top=3,
                align_x=0.5,
                align_y=0.0,
            )

    return DefaultConnectionLostComponent()


@final
class App:
    """
    Contains all the information needed to run a Rio app.

    Apps group all the information needed for Rio to run your application, such
    as its name, icon and, and the pages it contains. Apps also expose several
    lifetime events that you can use to perform tasks such as initialization and
    cleanup.

    If you're serving your app as a website, all users share the same `App`
    instance. If running in a window, there's only one window, and thus `App`,
    anyway.

    A basic setup may look like this:

    ```py
    app = rio.App(
        name="My App",
        build=MyAppRoot,
    )
    ```

    You can then run this app, either as a local application in a window:

    ```py
    app.run_in_window()
    ```

    Or you can create and run a webserver:

    ```py
    app.run_as_web_server()
    ```

    Or create a server, without running it. This allows you to start the script
    externally with tools such as uvicorn:

    ```py
    fastapi_app = app.as_fastapi()
    ```


    ## Attributes

    `name`: The name to display for this app. This can show up in window
        titles, error messages and wherever else the app needs to be
        referenced in a nice, human-readable way.

    `pages`: The pages that make up this app. You can navigate between these
        using `Session.navigate_to` or using `Link` components. If running
        as website the user can also access these pages directly via their
        URL.

    `assets_dir`: The directory where the app's assets are stored. This allows
        you to conveniently access any images or other files that are needed
        by your app.
    """

    # Type hints so the documentation generator knows which fields exist
    name: str
    pages: tuple[rio.Page, ...]
    assets_dir: Path

    def __init__(
        self,
        *,
        name: str | None = None,
        build: Callable[[], rio.Component] | None = None,
        icon: ImageLike | None = None,
        pages: Iterable[rio.Page] = tuple(),
        on_app_start: rio.EventHandler[App] = None,
        on_app_close: rio.EventHandler[App] = None,
        on_session_start: rio.EventHandler[rio.Session] = None,
        on_session_close: rio.EventHandler[rio.Session] = None,
        default_attachments: Iterable[Any] = (),
        ping_pong_interval: int | float | timedelta = timedelta(seconds=50),
        assets_dir: str | os.PathLike = "assets",
        theme: rio.Theme | tuple[rio.Theme, rio.Theme] | None = None,
        build_connection_lost_message: Callable[
            [], rio.Component
        ] = make_default_connection_lost_component,
    ):
        """
        ## Parameters

        `build`: A function that returns the root component of the app. This
            function will be called whenever a new session is created. Note
            that since classes are callable in Python, you can pass a class
            here instead of a function, so long as the class doesn't require
            any arguments.

            If no build method is passed, the app will create a `PageView`
            as the root component.

        `name`: The name to display for this app. This can show up in window
            titles, error messages and wherever else the app needs to be
            referenced in a nice, human-readable way. If not specified,
            `Rio` name will try to guess a name based on the name of the
            main Python file.

        `icon`: The icon to display for this app. This can show up in window
            the title bars of windows, browser tabs, or similar.

        `theme`: The `Theme` for the app. You can also pass in a tuple of two
            themes, which will be used as the light mode theme and the dark mode
            theme.

        `pages`: The pages that make up this app. You can navigate between
            these using `Session.navigate_to` or using `Link` components. If
            running as website the user can also access these pages directly
            via their URL.

        `on_app_start`: A function that will be called when the app is first
            started. You can use this to perform any initialization tasks
            that need to happen before the app is ready to use.

            The app start will be delayed until this function returns. This
            makes sure initialization is complete before the app is
            displayed to the user. If you would prefer to perform
            initialization in the background try using `asyncio.create_task`
            to run your code in a separate task.

        `on_app_close`: A function that will be called right before the app
            shuts down. You can use this to clean up open resources like for
            example a database connection.

        `on_session_start`: A function that will be called each time a new
            session is created. In the context of a website that would be
            each time a new user visits the site. In the context of a window
            there is only one session, so this will only be called once.

            This function does not block the creation of the session. This
            is to make sure initialization code doesn't accidentally make
            the user wait.

        `on_session_close`: A function that will be called each time a session
            ends. In the context of a website that would be each time a user
            closes their browser tab. In the context of a window this will
            only be called once, when the window is closed.

        `default_attachments`: A list of attachments that will be attached to
            every new session.

        `ping_pong_interval`: Rio periodically sends ping-pong messages
            between the client and server to prevent overzealous proxies
            from closing the connection. The default value should be fine
            for most deployments, but feel free to change it if your hosting
            provider deploys a particularly obnoxious proxy.

        `assets_dir`: The directory where the app's assets are stored. This
            allows you to conveniently access any images or other files that
            are needed by your app. If not specified, Rio will assume the
            assets are stored in a directory called "assets" in the same
            directory as the main Python file.

        `build_connection_lost_message`: A function that creates a "Connection
            lost" error popup, in case you want to override the default one.
        """
        main_file = _get_main_file()

        if name is None:
            name = _get_default_app_name(main_file)

        if icon is None:
            icon = utils.HOSTED_ASSETS_DIR / "rio-logos/rio-logo-square.png"

        if build is None:
            build = rio.PageView

        if theme is None:
            theme = rio.Theme.from_colors()

        # The `main_file` isn't detected correctly if the app is launched via
        # `rio run`. We'll store the user input so that `rio run` can fix the
        # assets dir.
        self._assets_dir = assets_dir
        self.assets_dir = main_file.parent / assets_dir

        self.name = name
        self._build = build
        self._icon = assets.Asset.from_image(icon)
        self.pages = tuple(pages)
        self._on_app_start = on_app_start
        self._on_app_close = on_app_close
        self._on_session_start = on_session_start
        self._on_session_close = on_session_close
        self.default_attachments: MutableSequence[Any] = list(
            default_attachments
        )
        self._theme = theme
        self._build_connection_lost_message = build_connection_lost_message

        if isinstance(ping_pong_interval, timedelta):
            self._ping_pong_interval = ping_pong_interval
        else:
            self._ping_pong_interval = timedelta(seconds=ping_pong_interval)

    def _as_fastapi(
        self,
        *,
        debug_mode: bool,
        running_in_window: bool,
        validator_factory: Callable[[rio.Session], debug.Validator] | None,
        internal_on_app_start: Callable[[], Any] | None,
    ) -> fastapi.FastAPI:
        """
        Internal equivalent of `as_fastapi` that takes additional arguments.
        """
        # Make sure all globals are initialized. This should be done as late as
        # possible, because it depends on which modules have been imported into
        # `sys.modules`.
        maybes.initialize()

        # Build the fastapi instance
        return app_server.AppServer(
            self,
            debug_mode=debug_mode,
            running_in_window=running_in_window,
            validator_factory=validator_factory,
            internal_on_app_start=internal_on_app_start,
        )

    def as_fastapi(self) -> fastapi.FastAPI:
        """
        Return a FastAPI instance that serves this app.

        This method returns a FastAPI instance that serves this app. This allows
        you to run the app with a custom server, such as uvicorn:

        ```py
        app = rio.App(
            name="My App",
            build=MyAppRoot,
        )

        fastapi_app = app.as_fastapi()
        ```

        You can then run this app with uvicorn:

        ```sh
        uvicorn my_app:fastapi_app
        ```
        """
        return self._as_fastapi(
            debug_mode=False,
            running_in_window=False,
            validator_factory=None,
            internal_on_app_start=None,
        )

    def _run_as_web_server(
        self,
        *,
        host: str,
        port: int,
        quiet: bool,
        running_in_window: bool,
        validator_factory: Callable[[rio.Session], debug.Validator]
        | None = None,
        internal_on_app_start: Callable[[], None] | None = None,
        internal_on_server_created: Callable[[uvicorn.Server], None]
        | None = None,
    ) -> None:
        """
        Internal equivalent of `run_as_web_server` that takes additional
        arguments.
        """
        port = utils.ensure_valid_port(host, port)

        # Suppress stdout messages if requested
        kwargs = {}

        if quiet:
            kwargs["log_config"] = {
                "version": 1,
                "disable_existing_loggers": True,
                "formatters": {},
                "handlers": {},
                "loggers": {},
            }

        # Create the FastAPI server
        fastapi_app = self._as_fastapi(
            debug_mode=False,
            running_in_window=running_in_window,
            validator_factory=validator_factory,
            internal_on_app_start=internal_on_app_start,
        )

        # Suppress stdout messages if requested
        log_level = "error" if quiet else "info"

        config = uvicorn.Config(
            fastapi_app,
            host=host,
            port=port,
            log_level=log_level,
            timeout_graceful_shutdown=1,  # Without a timeout, sometimes the server just deadlocks
        )
        server = uvicorn.Server(config)

        if internal_on_server_created is not None:
            internal_on_server_created(server)

        server.run()

    def run_as_web_server(
        self,
        *,
        host: str = "localhost",
        port: int = 8000,
        quiet: bool = False,
    ) -> None:
        """
        Creates and runs a webserver that serves this app.

        This method creates and immediately runs a webserver that serves this
        app. This is the simplest way to run a Rio app.

        ```py
        app = rio.App(
            name="My App",
            build=MyAppRoot,
        )

        app.run_as_web_server()
        ```

        The will synchronously block until the server is shut down.

        ## Parameters

        host: Which IP address to serve the webserver on. `localhost` will
            make the service only available on your local machine. This is
            the recommended setting if running behind a proxy like nginx.

        port: Which port the webserver should listen to.

        quiet: If `True` Rio won't send any routine messages to `stdout`.
            Error messages will be printed regardless of this setting.
        """
        self._run_as_web_server(
            host=host,
            port=port,
            quiet=quiet,
            running_in_window=False,
        )

    def run_in_browser(
        self,
        *,
        host: str = "localhost",
        port: int | None = None,
        quiet: bool = False,
    ) -> None:
        """
        Runs an internal webserver and opens the app in the default browser.

        This method creates and immediately runs a webserver that serves this
        app, and then opens the app in the default browser. This is a quick and
        easy way to access your app.

        ```py
        app = rio.App(
            name="My App",
            build=MyAppRoot,
        )

        app.run_in_browser()
        ```

        ## Parameters
        host: Which IP address to serve the webserver on. `localhost` will
            make the service only available on your local machine. This is the
            recommended setting if running behind a proxy like nginx.

        port: Which port the webserver should listen to. If not specified,
            Rio will choose a random free port.

        quiet: If `True` Rio won't send any routine messages to `stdout`.
            Error messages will be printed regardless of this setting.
        """
        port = utils.ensure_valid_port(host, port)

        def on_startup() -> None:
            webbrowser.open(f"http://{host}:{port}")

        self._run_as_web_server(
            host=host,
            port=port,
            quiet=quiet,
            running_in_window=False,
            internal_on_app_start=on_startup,
        )

    def run_in_window(
        self,
        quiet: bool = True,
    ) -> None:
        """
        Runs the app in a local window.

        This method creates a window and displays the app in it. This is great
        if you don't want the complexity of running a web server, or wish to
        package your app as a standalone executable.

        ```py
        app = rio.App(
            name="My App",
            build=MyAppRoot,
        )

        app.run_in_window()
        ```

        This method requires the `window` extra. If you don't have it installed,
        you can install it with:

        ```sh
        pip install rio-ui[window]
        ```

        This method will synchronously block until the window is closed.


        ## Parameters

        `quiet`: If `True` Rio won't send any routine messages to `stdout`.
            Error messages will be printed regardless of this setting.
        """
        try:
            import webview  # type: ignore
        except ImportError:
            raise Exception(
                "The `window` extra is required to use `App.run_in_window`."
                " Run `pip install rio-ui[window]` to install it."
            ) from None

        # Unfortunately, WebView must run in the main thread, which makes this
        # tricky. We'll have to banish uvicorn to a background thread, and shut
        # it down when the window is closed.

        host = "localhost"
        port = utils.ensure_valid_port(host, None)
        url = f"http://{host}:{port}"

        # This lock is released once the server is running
        app_ready_event = threading.Event()

        server: uvicorn.Server | None = None

        def on_server_created(serv: uvicorn.Server) -> None:
            nonlocal server
            server = serv

        # Start the server, and release the lock once it's running
        def run_web_server():
            self._run_as_web_server(
                host=host,
                port=port,
                quiet=quiet,
                running_in_window=True,
                internal_on_app_start=app_ready_event.set,
                internal_on_server_created=on_server_created,
            )

        server_thread = threading.Thread(target=run_web_server)
        server_thread.start()

        # Wait for the server to start
        app_ready_event.wait()

        # Start the webview
        try:
            webview.create_window(self.name, url)
            webview.start(debug=os.environ.get("RIO_WEBVIEW_DEBUG") == "1")

        finally:
            assert isinstance(server, uvicorn.Server)

            server.should_exit = True
            server_thread.join()


def _get_main_file() -> Path:
    try:
        main_file = Path(__main__.__file__)
    except AttributeError:
        main_file = Path(sys.argv[0])

    # Find out if we're being executed by uvicorn
    if (
        main_file.name != "__main__.py"
        or main_file.parent != Path(uvicorn.__file__).parent
    ):
        return main_file

    # Find out from which module uvicorn imported the app
    try:
        app_location = next(arg for arg in sys.argv[1:] if ":" in arg)
    except StopIteration:
        return main_file

    module_name, _, _ = app_location.partition(":")
    module = sys.modules[module_name]

    if module.__file__ is None:
        return main_file

    return Path(module.__file__)


def _get_default_app_name(main_file: Path) -> str:
    name = main_file.stem
    if name in ("main", "__main__", "__init__"):
        name = main_file.absolute().parent.stem

    return name.replace("_", " ").title()



import dataclasses

from utils import enable_component_instantiation

import rio.testing


@enable_component_instantiation
def test_fields_with_defaults():
    class TestComponent(rio.Component):
        foo: list[str] = dataclasses.field(init=False, default_factory=list)
        bar: int = dataclasses.field(init=False, default=5)

        def build(self) -> rio.Component:
            raise NotImplementedError()

    component = TestComponent()
    assert component.foo == []
    assert component.bar == 5


async def test_post_init():
    class TestComponent(rio.Component):
        post_init_called: bool = False

        def __post_init__(self):
            self.post_init_called = True

        def build(self) -> rio.Component:
            return rio.Text("hi")

    async with rio.testing.TestClient(TestComponent) as test_client:
        root_component = test_client.get_component(TestComponent)
        assert root_component.post_init_called



from typing import final

from .fundamental_component import FundamentalComponent

__all__ = ["Html"]


@final
class Html(FundamentalComponent):
    """
    Renders HTML.

    `Html` allows you to embed HTML in your app. It takes HTML code as input and
    renders it. If you want to embed an entire website, use the `Website`
    component instead.


    ## Attributes

    `html`: The HTML to render.


    ## Examples

    This minimal example will render a simple HTML code:

    ```python
    rio.Html('''
    <h1>Hello World</h1>
    <p>Welcome to rio!</p>
    ''')
    """

    html: str


Html._unique_id = "Html-builtin"



from dataclasses import dataclass
from pathlib import Path


class Event:
    pass


@dataclass
class FileChanged(Event):
    """
    A file in the project has changed, necessitating a reload.
    """

    timestamp_nanoseconds: (
        int  # Monotonic timestamp of the change, in nanoseconds
    )
    path_to_file: Path


@dataclass
class StopRequested(Event):
    """
    Request a shutdown of the running project.
    """

    pass



from __future__ import annotations

from typing import Literal, final

from uniserde import JsonDoc

import rio

from .fundamental_component import FundamentalComponent

__all__ = [
    "Link",
]


@final
class Link(FundamentalComponent):
    """
    Navigates to a page or URL when clicked.

    `Link`s display a short text, or arbitrary component, and navigate to a page
    or URL when clicked.


    ## Attributes

    `child_text`: The text to display inside the link.

    `child_component`: The component to display inside the link.

    `target_url`: The page or URL to navigate to when clicked.

    `open_in_new_tab`: Whether to open the link in a new tab. Defaults to `False`.


    ## Examples

    This minimal example will simply display a link with the URL
    "https://example.com" and a text "Click me!":

    ```python
    rio.Link("Click me!", "https://example.com")
    ```
    """

    # Exactly one of these will be set, the other `None`
    child_text: str | None
    child_component: rio.Component | None
    target_url: rio.URL | str
    open_in_new_tab: bool

    # The serializer can't handle Union types. Override the constructor, so it
    # splits the child into two values
    def __init__(
        self,
        content: rio.Component | str,
        target_url: rio.URL | str,
        *,
        open_in_new_tab: bool = False,
        key: str | None = None,
        margin: float | None = None,
        margin_x: float | None = None,
        margin_y: float | None = None,
        margin_left: float | None = None,
        margin_top: float | None = None,
        margin_right: float | None = None,
        margin_bottom: float | None = None,
        width: float | Literal["natural", "grow"] = "natural",
        height: float | Literal["natural", "grow"] = "natural",
        align_x: float | None = None,
        align_y: float | None = None,
    ):
        """
        ## Parameters

        `content`: The text or component to display inside the link.

        `target_url`: The page or URL to navigate to when clicked.
        """
        super().__init__(
            key=key,
            margin=margin,
            margin_x=margin_x,
            margin_y=margin_y,
            margin_left=margin_left,
            margin_top=margin_top,
            margin_right=margin_right,
            margin_bottom=margin_bottom,
            width=width,
            height=height,
            align_x=align_x,
            align_y=align_y,
        )

        if isinstance(content, str):
            self.child_text = content
            self.child_component = None
        else:
            self.child_text = None
            self.child_component = content

        self.target_url = target_url
        self.open_in_new_tab = open_in_new_tab

        self._properties_set_by_creator_.update(
            ("child_text", "child_component")
        )

    def _custom_serialize(self) -> JsonDoc:
        # Get the full URL to navigate to
        target_url_absolute = self.session.active_page_url.join(
            rio.URL(self.target_url)
        )

        # Serialize everything
        return {
            "targetUrl": str(target_url_absolute),
        }


Link._unique_id = "Link-builtin"



import asyncio
import socket
from typing import *  # type: ignore

import revel
import uvicorn
import uvicorn.lifespan.on

import rio
import rio.app_server
import rio.cli

from .. import nice_traceback
from . import run_models


class UvicornWorker:
    def __init__(
        self,
        *,
        push_event: Callable[[run_models.Event], None],
        app: rio.App,
        socket: socket.socket,
        quiet: bool,
        debug_mode: bool,
        run_in_window: bool,
        on_server_is_ready_or_failed: asyncio.Future[None],
    ) -> None:
        self.push_event = push_event
        self.app = app
        self.socket = socket
        self.quiet = quiet
        self.debug_mode = debug_mode
        self.run_in_window = run_in_window
        self.on_server_is_ready_or_failed = on_server_is_ready_or_failed

        # The app server used to host the app
        self.app_server: rio.app_server.AppServer | None = None

    async def run(self) -> None:
        rio.cli._logger.debug("Uvicorn worker is starting")

        # Set up a uvicorn server, but don't start it yet
        app_server = self.app._as_fastapi(
            debug_mode=self.debug_mode,
            running_in_window=self.run_in_window,
            validator_factory=None,
            internal_on_app_start=lambda: self.on_server_is_ready_or_failed.set_result(
                None
            ),
        )
        assert isinstance(app_server, rio.app_server.AppServer)
        self.app_server = app_server

        config = uvicorn.Config(
            self.app_server,
            log_config=None,  # Prevent uvicorn from configuring global logging
            log_level="error" if self.quiet else "info",
            timeout_graceful_shutdown=1,  # Without a timeout the server sometimes deadlocks
        )
        self._uvicorn_server = uvicorn.Server(config)

        # TODO: Uvicorn wishes to set up the event loop. The current code
        # doesn't let it do that, since asyncio is already running.
        #
        # self._uvicorn_server.config.setup_event_loop()

        # Uvicorn can't handle `CancelledError` properly, so make sure it never
        # sees one. Uvicorn is wrapped in a task, which can finish peacefully.
        serve_task = asyncio.create_task(
            self._uvicorn_server.serve(
                sockets=[self.socket],
            ),
            name="uvicorn serve",
        )

        # Run the server
        try:
            await asyncio.shield(serve_task)
        except asyncio.CancelledError:
            pass
        except Exception as err:
            rio.cli._logger.exception(f"Uvicorn has crashed")

            revel.error(f"Uvicorn has crashed:")
            print()
            revel.print(nice_traceback.format_exception_revel(err))
            self.push_event(run_models.StopRequested())
        finally:
            rio.cli._logger.debug("Requesting uvicorn to exit")

            self._uvicorn_server.should_exit = True

            # Make sure not to return before the task has returned, but also
            # avoid receiving any exceptions _again_
            try:
                await serve_task
            except:
                pass
            finally:
                rio.cli._logger.debug("Uvicorn serve task has ended")

    def replace_app(self, app: rio.App) -> None:
        """
        Replace the app currently running in the server with a new one. The
        worker must already be running for this to work.
        """
        assert self.app_server is not None
        rio.cli._logger.debug("Replacing the app in the server")
        self.app_server.app = app



from __future__ import annotations

import enum
from datetime import timedelta
from typing import *  # type: ignore

__all__ = [
    "on_page_change",
    "on_mount",
    "on_unmount",
    "periodic",
]


R = TypeVar("R")
SyncOrAsync = R | Awaitable[R]
SyncOrAsyncNone = TypeVar("SyncOrAsyncNone", bound=SyncOrAsync[None])

Func = TypeVar("Func", bound=Callable)
Decorator = Callable[[Func], Func]

MethodWithNoParametersVar = TypeVar(
    "MethodWithNoParametersVar", bound=Callable[[Any], Any]
)


class EventTag(enum.Enum):
    """
    Event tags are internal markers used to keep track of which function needs
    to be called when.

    ## Metadata

    public: False
    """

    ON_POPULATE = enum.auto()
    ON_PAGE_CHANGE = enum.auto()
    ON_MOUNT = enum.auto()
    ON_UNMOUNT = enum.auto()
    PERIODIC = enum.auto()


def _register_as_event_handler(
    function: Callable, tag: EventTag, args: Any
) -> None:
    all_events: dict[EventTag, list[Any]] = vars(function).setdefault(
        "_rio_events_", {}
    )
    events_like_this = all_events.setdefault(tag, [])
    events_like_this.append(args)


def on_populate(
    handler: MethodWithNoParametersVar,
) -> MethodWithNoParametersVar:
    """
    Triggered after the component has been created or has been reconciled. This
    allows you to asynchronously fetch any data which depends on the component's
    state.
    """
    _register_as_event_handler(handler, EventTag.ON_POPULATE, None)
    return handler


def on_page_change(
    handler: MethodWithNoParametersVar,
) -> MethodWithNoParametersVar:
    """
    Triggered whenever the session changes pages.
    """
    _register_as_event_handler(handler, EventTag.ON_PAGE_CHANGE, None)
    return handler


def on_mount(handler: MethodWithNoParametersVar) -> MethodWithNoParametersVar:
    """
    Triggered when the component is added to the component tree.

    This may be triggered multiple times if the component is removed and then
    re-added.
    """
    _register_as_event_handler(handler, EventTag.ON_MOUNT, None)
    return handler


def on_unmount(handler: MethodWithNoParametersVar) -> MethodWithNoParametersVar:
    """
    Triggered when the component is removed from the component tree.

    This may be triggered multiple times if the component is removed and then
    re-added.
    """
    _register_as_event_handler(handler, EventTag.ON_UNMOUNT, None)
    return handler


def periodic(
    interval: float | timedelta,
) -> Decorator[MethodWithNoParametersVar]:
    """
    This event is triggered repeatedly at a fixed time interval for as long as
    the component exists. The component does not have to be mounted for this
    event to trigger.

    The interval only starts counting after the previous handler has finished
    executing, so the handler will never run twice simultaneously, even if it
    takes longer than the interval to execute.


    ## Parameters

    `period`: The number of seconds, or timedelta, between each trigger.


    ## Metadata

    `experimental`: True
    """
    # Convert timedelta to float
    if isinstance(interval, timedelta):
        interval = interval.total_seconds()

    def decorator(
        handler: MethodWithNoParametersVar,
    ) -> MethodWithNoParametersVar:
        _register_as_event_handler(handler, EventTag.PERIODIC, interval)
        return handler

    return decorator



from typing import *  # type: ignore

import rio.components.component_tree

from . import component_details


class TreePage(rio.Component):
    _selected_component_id: int | None = None

    def _on_select_component(self, component_id: int) -> None:
        self._selected_component_id = component_id

    def build(self) -> rio.Component:
        margin = 1

        children = [
            rio.Text(
                "Component Tree",
                style="heading2",
                margin=margin,
                justify="left",
            ),
            rio.components.component_tree.ComponentTree(
                width=10,
                height="grow",
                margin_left=margin,
                on_select_component=self._on_select_component,
            ),
        ]

        if self._selected_component_id is not None:
            children.append(
                component_details.ComponentDetails(
                    component_id=self._selected_component_id,
                    margin=margin,
                )
            )

        return rio.Column(*children)



import asyncio
import json
import types
from collections.abc import Callable, Iterable, Iterator, Mapping
from typing_extensions import Any, Self, TypeVar, overload

import ordered_set
from uniserde import Jsonable, JsonDoc

import rio
from rio.app_server import AppServer
from rio.components.root_components import (
    FundamentalRootComponent,
    HighLevelRootComponent,
)


__all__ = ["TestClient"]


T = TypeVar("T")
C = TypeVar("C", bound=rio.Component)


class TestClient:
    @overload
    def __init__(
        self,
        app: rio.App,
        *,
        running_in_window: bool = False,
        user_settings: JsonDoc = {},
        active_url: str = "/",
        use_ordered_dirty_set: bool = False,
    ): ...

    @overload
    def __init__(
        self,
        build: Callable[[], rio.Component] = rio.Spacer,
        *,
        app_name: str = "mock-app",
        default_attachments: Iterable[object] = (),
        running_in_window: bool = False,
        user_settings: JsonDoc = {},
        active_url: str = "/",
        use_ordered_dirty_set: bool = False,
    ): ...

    def __init__(  # type: ignore
        self,
        app_or_build: rio.App | Callable[[], rio.Component] | None = None,
        *,
        app: rio.App | None = None,
        build: Callable[[], rio.Component] | None = None,
        app_name: str = "mock-app",
        default_attachments: Iterable[object] = (),
        running_in_window: bool = False,
        user_settings: JsonDoc = {},
        active_url: str = "/",
        use_ordered_dirty_set: bool = False,
    ):
        if app is None:
            if isinstance(app_or_build, rio.App):
                app = app_or_build
            else:
                if build is None:
                    if app_or_build is not None:
                        build = app_or_build
                    else:
                        build = rio.Spacer

                app = rio.App(
                    build=build,
                    name=app_name,
                    default_attachments=tuple(default_attachments),
                )

        self._app_server = AppServer(
            app,
            debug_mode=False,
            running_in_window=running_in_window,
            validator_factory=None,
            internal_on_app_start=None,
        )

        self._use_ordered_dirty_set = use_ordered_dirty_set

        self._session: rio.Session | None = None
        self._server_task: asyncio.Task | None = None
        self._outgoing_messages = list[JsonDoc]()
        self._responses = asyncio.Queue[JsonDoc]()
        self._responses.put_nowait(
            {
                "websiteUrl": "https://unit.test" + active_url,
                "preferredLanguages": [],
                "userSettings": user_settings,
                "windowWidth": 1920,
                "windowHeight": 1080,
                "timezone": "America/New_York",
                "decimalSeparator": ".",
                "thousandsSeparator": ",",
                "prefersLightTheme": True,
            }
        )
        self._first_refresh_completed = asyncio.Event()

    async def _send_message(self, message_text: str) -> None:
        message = json.loads(message_text)

        self._outgoing_messages.append(message)

        if "id" in message:
            self._responses.put_nowait(
                {
                    "jsonrpc": "2.0",
                    "id": message["id"],
                    "result": None,
                }
            )

        if message["method"] == "updateComponentStates":
            self._first_refresh_completed.set()

    async def _receive_message(self) -> Jsonable:
        return await self._responses.get()

    async def __aenter__(self) -> Self:
        # Emulate the process of creating a session as closely as possible
        fake_request: Any = types.SimpleNamespace(
            url="https://unit.test",
            base_url="https://unit.test",
            headers={"accept": "text/html"},
            client=types.SimpleNamespace(host="localhost", port="12345"),
        )
        await self._app_server._serve_index(fake_request, "")

        [[session_token, session]] = (
            self._app_server._active_session_tokens.items()
        )
        self._session = session

        if self._use_ordered_dirty_set:
            session._dirty_components = ordered_set.OrderedSet(
                session._dirty_components
            )  # type: ignore

        fake_websocket: Any = types.SimpleNamespace(
            client="1.2.3.4",
            accept=lambda: _make_awaitable(),
            send_text=self._send_message,
            receive_json=self._receive_message,
        )

        test_task = asyncio.current_task()
        assert test_task is not None

        async def serve_websocket():
            try:
                await self._app_server._serve_websocket(
                    fake_websocket, session_token
                )
            except asyncio.CancelledError:
                pass
            except Exception as error:
                test_task.cancel(
                    f"Exception in AppServer._serve_websocket: {error}"
                )
            else:
                test_task.cancel(
                    "AppServer._serve_websocket exited unexpectedly. An"
                    " exception must have occurred in the `init_coro`."
                )

        self._server_task = asyncio.create_task(serve_websocket())

        await self._first_refresh_completed.wait()
        return self

    async def __aexit__(self, *_) -> None:
        if self._server_task is not None:
            self._server_task.cancel()

    @property
    def _dirty_components(self) -> set[rio.Component]:
        return set(self.session._dirty_components)

    @property
    def _last_updated_components(self) -> set[rio.Component]:
        return set(self._last_component_state_changes)

    @property
    def _last_component_state_changes(
        self,
    ) -> Mapping[rio.Component, Mapping[str, object]]:
        for message in reversed(self._outgoing_messages):
            if message["method"] == "updateComponentStates":
                delta_states: dict = message["params"]["deltaStates"]  # type: ignore
                return {
                    self.session._weak_components_by_id[
                        int(component_id)
                    ]: delta
                    for component_id, delta in delta_states.items()
                    if int(component_id) != self.session._root_component._id
                }

        return {}

    def _get_build_output(
        self,
        component: rio.Component,
        type_: type[C] | None = None,
    ) -> C:
        result = self.session._weak_component_data_by_component[
            component
        ].build_result

        if type_ is not None:
            assert (
                type(result) is type_
            ), f"Expected {type_}, got {type(result)}"

        return result  # type: ignore

    @property
    def session(self) -> rio.Session:
        assert self._session is not None

        return self._session

    @property
    def crashed_build_functions(self) -> Mapping[Callable, str]:
        return self.session._crashed_build_functions

    @property
    def root_component(self) -> rio.Component:
        high_level_root = self.session._root_component
        assert isinstance(
            high_level_root, HighLevelRootComponent
        ), high_level_root

        low_level_root = self.session._weak_component_data_by_component[
            high_level_root
        ].build_result
        assert isinstance(
            low_level_root, FundamentalRootComponent
        ), low_level_root

        scroll_container = low_level_root.content
        assert isinstance(
            scroll_container, rio.ScrollContainer
        ), scroll_container

        return scroll_container.content

    def get_components(self, component_type: type[C]) -> Iterator[C]:
        root_component = self.root_component

        for component in root_component._iter_component_tree():
            if type(component) is component_type:
                yield component  # type: ignore

    def get_component(self, component_type: type[C]) -> C:
        try:
            return next(self.get_components(component_type))
        except StopIteration:
            raise AssertionError(f"No component of type {component_type} found")

    async def refresh(self) -> None:
        await self.session._refresh()


async def _make_awaitable(value: T = None) -> T:
    return value



import re
import subprocess
import sys
import tempfile
import textwrap
from typing import *  # type: ignore

import black
import pyright
import pytest

import rio.docs

CODE_BLOCK_PATTERN = re.compile(r"```(.*?)```", re.DOTALL)


all_documented_objects = [
    obj for obj, _ in rio.docs.find_documented_objects(False)
]
all_documented_objects.sort(key=lambda obj: obj.__name__)


def get_code_blocks(obj: type | Callable) -> list[str]:
    """
    Returns a list of all code blocks in the docstring of a component.
    """
    docstring = obj.__doc__

    # No docs?
    if not docstring:
        return []

    docstring = textwrap.dedent(docstring)

    # Find any contained code blocks
    result: list[str] = []
    for match in CODE_BLOCK_PATTERN.finditer(docstring):
        block: str = match.group(1)

        # Split into language and source
        linebreak = block.find("\n")
        assert linebreak != -1
        language = block[:linebreak]
        block = block[linebreak + 1 :]

        # Make sure a language is specified
        assert language, "The code block has no language specified"

        result.append(block)

    return result


@pytest.mark.parametrize("obj", all_documented_objects)
def test_eval_code_block(obj: type | Callable) -> None:
    # Eval all code blocks and make sure they don't crash
    for source in get_code_blocks(obj):
        compile(source, "<string>", "exec")


@pytest.mark.parametrize("obj", all_documented_objects)
def test_code_block_is_formatted(obj: type | Callable) -> None:
    # Make sure all code blocks are formatted according to black
    for source in get_code_blocks(obj):
        formatted_source = black.format_str(source, mode=black.FileMode())

        # Black often inserts 2 empty lines between stuff, but that's really not
        # necessary in docstrings. So we'll collapse those into a single empty
        # line.
        source = source.replace("\n\n\n", "\n\n")
        formatted_source = formatted_source.replace("\n\n\n", "\n\n")

        assert source == formatted_source


PYRIGHT_ERROR_OR_WARNING_REGEX = re.compile(
    r".*\.py:\d+:\d+ - (?:error|warning): (.*)"
)


def _find_static_typing_errors(source: str) -> str:
    """
    Run pyright on the given source and return the number of errors and
    warnings.
    """
    with tempfile.NamedTemporaryFile(suffix=".py") as f:
        # Dump the source to a file, and implicitly import rio
        f.write("import rio\n".encode("utf-8"))
        f.write(source.encode("utf-8"))
        f.flush()

        # Run pyright
        proc = pyright.run(
            "--pythonpath",
            sys.executable,
            f.name,
            stdout=subprocess.PIPE,
            stderr=subprocess.PIPE,
        )
        result_out = proc.stdout
        assert isinstance(result_out, bytes), type(result_out)
        result_out = result_out.decode()

        # Find the number of errors and warnings
        lines = list[str]()

        for line in result_out.splitlines():
            match = PYRIGHT_ERROR_OR_WARNING_REGEX.match(line)

            if match:
                lines.append(match.group(1))

        return "\n".join(lines)


@pytest.mark.parametrize("obj", all_documented_objects)
def test_analyze_code_block(obj: type | Callable) -> None:
    # A lot of snippets are missing context, so it's only natural that pyright
    # will find issues with the code. There isn't really anything we can do
    # about it, so we'll just skip those object.
    if obj in (rio.App, rio.Color, rio.UserSettings):
        pytest.xfail()

    # Make sure pyright is happy with all code blocks
    for source in get_code_blocks(obj):
        errors = _find_static_typing_errors(source)
        assert not errors, errors



import rio


# <component>
class SamplePage(rio.Component):
    """
    This is an example Page. Pages are identical to other Components and only
    differ in how they're used.
    """

    def build(self) -> rio.Component:
        return rio.Column(
            rio.Text("My App", style="heading2"),
            rio.Text(
                "This is a sample page. Replace it with your own content."
            ),
            spacing=2,
            margin=2,
            align_x=0.5,
            align_y=0,
        )


# </component>



import io
import re
import shutil
import string
from pathlib import Path
from typing import *  # type: ignore

import introspection
import isort
import revel
from revel import fatal, print, success

import rio.cli
import rio.snippets

__all__ = [
    "create_project",
]


def class_name_from_snippet(snip: rio.snippets.Snippet) -> str:
    """
    Given a file name, determine the name of the class that is defined in it.

    e.g. `sample_component.py` -> `SampleComponent`
    """
    assert snip.name.endswith(".py"), snip.name

    parts = snip.name[:-3].split("_")
    return "".join(part.capitalize() for part in parts)


def write_init_file(fil: IO, snippets: Iterable[rio.snippets.Snippet]) -> None:
    """
    Write an `__init__.py` file that imports all of the snippets.

    e.g. if told to import snippets `foo.py` and `bar.py`, it will write:

    ```
    from .foo import Foo
    from .bar import Bar
    ```
    """
    for snippet in snippets:
        assert snippet.name.endswith(".py"), snippet.name
        module_name = snippet.name[:-3]
        class_name = class_name_from_snippet(snippet)
        fil.write(f"from .{module_name} import {class_name}\n")


def generate_root_init(
    out: TextIO,
    *,
    raw_name: str,
    project_type: Literal["app", "website"],
    components: List[rio.snippets.Snippet],
    pages: List[rio.snippets.Snippet],
    homepage_snippet: rio.snippets.Snippet,
    root_init_snippet: rio.snippets.Snippet,
    on_app_start: str | None = None,
    default_attachments: list[str] | None = None,
) -> None:
    """
    Generate the `__init__.py` file for the main module of the project.
    """
    assert len(pages) > 0, pages

    # If a page is called `root_page.py`, it will be used as the root component
    root_page_snippet: rio.snippets.Snippet | None = None

    for page in pages:
        if page.name == "root_page.py":
            root_page_snippet = page
            break

    # Ensure the homepage_snippet is the first in the pages list
    if pages[0] != homepage_snippet:
        pages = [homepage_snippet] + [
            page for page in pages if page != homepage_snippet
        ]

    # Prepare the different pages
    page_strings = []
    for snip in pages:
        page_component_name = class_name_from_snippet(snip)

        # Don't add the root page to the list of pages
        if snip is root_page_snippet:
            continue

        # What's the URL segment for this page?
        if snip is homepage_snippet:
            url_segment = ""
            page_nicename = "Home"
        else:
            assert snip.name.endswith(".py"), snip.name
            url_segment = snip.name[:-3].replace("_", "-").lower()
            page_nicename = page_component_name

        page_strings.append(
            f"""
        rio.Page(
            name="{page_nicename}",
            page_url={url_segment!r},
            build=pages.{page_component_name},
        ),"""
        )

    page_string = "\n".join(page_strings)

    # Prepare the default attachments
    if default_attachments is None:
        default_attachment_string = ""
    else:
        default_attachment_string = (
            f"\n    default_attachments=[{', '.join(default_attachments)}],"
        )

    # Imports
    default_theme = rio.Theme.from_colors()
    buffer = io.StringIO()

    buffer.write(
        f"""
from __future__ import annotations

from pathlib import Path
from typing import *  # type: ignore

import rio

from . import pages
from . import components as comps
    """.strip()
    )

    # Additional imports
    try:
        additional_imports = root_init_snippet.get_section("additional-imports")
    except KeyError:
        needs_isort = False
    else:
        buffer.write("\n")
        buffer.write(additional_imports)
        buffer.write("\n\n")
        needs_isort = True

    # Additional code
    try:
        additional_code = root_init_snippet.get_section("additional-code")
    except KeyError:
        pass
    else:
        buffer.write("\n")
        buffer.write(additional_code)
        buffer.write("\n\n")

    # Theme & App generation
    buffer.write(
        f"""

# Define a theme for Rio to use.
#
# You can modify the colors here to adapt the appearance of your app or website.
# The most important parameters are listed, but more are available! You can find
# them all in the docs
#
# https://rio.dev/docs/api/theme
theme = rio.Theme.from_colors(
    primary_color=rio.Color.from_hex("{default_theme.primary_color.hex}"),
    secondary_color=rio.Color.from_hex("{default_theme.secondary_color.hex}"),
    light=True,
)


# Create the Rio app
app = rio.App(
    name={raw_name!r},
    pages=[{page_string}
    ],{default_attachment_string}
"""
    )

    # Some parameters are optional
    if on_app_start is not None:
        buffer.write(
            f"    # This function will be called once the app is ready.\n"
            f"    #\n"
            f"    # `rio run` will also call it again each time the app is reloaded.\n"
            f"    on_app_start={on_app_start},\n"
        )

    if root_page_snippet is not None:
        buffer.write(
            "    # You can optionally provide a root component for the app. By default,\n"
            "    # a simple `rio.PageView` is used. By providing your own component, you\n"
            "    # can create components which stay put while the user navigates between\n"
            "    # pages, such as a navigation bar or footer.\n"
            "    #\n"
            "    # When you do this, make sure your component contains a `rio.PageView`\n"
            "    # so the currently active page is still visible.\n"
            "    build=pages.RootPage,\n"
        )

    buffer.write("    theme=theme,\n")
    buffer.write('    assets_dir=Path(__file__).parent / "assets",\n')
    buffer.write(")\n\n")

    # Due to imports coming from different sources they're often not sorted.
    # -> Apply `isort`
    if needs_isort:
        formatted_code = isort.code(buffer.getvalue())
        out.write(formatted_code)
    else:
        out.write(buffer.getvalue())


def strip_invalid_filename_characters(name: str) -> str:
    """
    Given a name, strip any characters that are not allowed in a filename.
    """
    return re.sub(r'[<>:"/\\|?*]', "", name)


def derive_module_name(raw_name: str) -> str:
    """
    Given an arbitrary string, derive similar and valid all_lower Python module
    identifier from it.
    """
    # Convert to lower_case
    name = introspection.convert_case(raw_name, "snake")

    # Strip any invalid characters
    name = "".join(c for c in name if c.isidentifier() or c in string.digits)

    # Since modules are written to files, the name also has to be a valid file
    # name
    name = strip_invalid_filename_characters(name)

    # Identifiers cannot start with a digit
    while name and name[0].isdigit():
        name = name[1:]

    # This could've resulted in an empty string
    if not name:
        name = "rio_app"

    # Done
    return name


def generate_readme(
    out: TextIO,
    raw_name: str,
    template: rio.snippets.ProjectTemplate,
) -> None:
    out.write(
        f"""# {raw_name}

This is a placeholder README for your project. Use it to describe what your
project is about, to give new users a quick overview of what they can expect.

_{raw_name.capitalize()}_ was created using [Rio](http://rio.dev/), an easy to
use app & website framework for Python._
"""
    )

    # Include the template's README
    if template.name != "Empty":
        out.write(
            f"""
This project is based on the `{template.name}` template.

## {template.name}

{template.description_markdown_source}
"""
        )


def write_component_file(
    out: TextIO,
    snip: rio.snippets.Snippet,
) -> None:
    """
    Writes the Python file containing a component or page to the given file.
    """
    # Common imports
    buffer = io.StringIO()

    buffer.write(
        f"""from __future__ import annotations

from dataclasses import KW_ONLY, field
from typing import *  # type: ignore

import rio

from .. import components as comps

"""
    )

    # Additional, user-defined imports
    try:
        additional_imports = snip.get_section("additional-imports")
    except KeyError:
        needs_isort = False
    else:
        buffer.write(additional_imports)
        buffer.write("\n")
        needs_isort = True

    # The component proper
    buffer.write(snip.get_section("component"))

    # Due to imports coming from different sources they're often not sorted.
    # -> Apply `isort`
    if needs_isort:
        formatted_code = isort.code(buffer.getvalue())
        out.write(formatted_code)
    else:
        out.write(buffer.getvalue())


def generate_dependencies_file(
    project_dir: Path, dependencies: dict[str, str]
) -> None:
    """
    Writes a `requirements.txt` file with the given dependencies. Does nothing
    if there are no dependencies.
    """
    # Anything to do?
    if not dependencies:
        return

    # requirements.txt
    with open(project_dir / "requirements.txt", "w", encoding="utf-8") as out:
        for package, version_specifier in dependencies.items():
            out.write(f"{package}{version_specifier}\n")


def create_project(
    *,
    raw_name: str,
    type: Literal["app", "website"],
    template_name: rio.snippets.AvailableTemplatesLiteral,
) -> None:
    """
    Create a new project with the given name. This will directly interact with
    the terminal, asking for input and printing output.
    """

    # Derive a valid module name
    module_name = derive_module_name(raw_name)

    # The project directory is called the same, but in kebab-case
    dashed_name = module_name.replace("_", "-")

    # Find the template
    for template in rio.snippets.get_project_templates(include_empty=True):
        if template.name == template_name:
            break
    else:
        assert False, f"Received invalid template name `{template_name}`. This shouldn't be possible if the types are correct."

    # Create the target directory
    project_dir = Path.cwd() / dashed_name
    project_dir.mkdir(parents=True, exist_ok=True)

    # If the project directory already exists it must be empty
    if any(project_dir.iterdir()):
        fatal(
            f"The project directory `{project_dir}` already exists and is not empty"
        )

    # Generate /rio.toml
    with open(project_dir / "rio.toml", "w", encoding="utf-8") as f:
        f.write("# This is the configuration file for Rio,\n")
        f.write("# an easy to use app & web framework for Python.\n")
        f.write("\n")
        f.write(f"[app]\n")
        f.write(f'app_type = "{type}"  # This is either "website" or "app"\n')
        f.write(
            f'main_module = "{module_name}"  # The name of your Python module\n'
        )

    # Create the main module and its subdirectories
    main_module_dir = project_dir / module_name
    assets_dir = main_module_dir / "assets"
    components_dir = main_module_dir / "components"
    pages_dir = main_module_dir / "pages"

    main_module_dir.mkdir(parents=True, exist_ok=True)
    assets_dir.mkdir()
    components_dir.mkdir()
    pages_dir.mkdir()

    # Generate /assets/*
    for snip in template.asset_snippets:
        source_path = snip.file_path
        target_path = assets_dir / source_path.name
        shutil.copyfile(source_path, target_path)

    # Generate /components/*.py
    for snip in template.component_snippets:
        target_path = components_dir / snip.name

        with target_path.open("w", encoding="utf-8") as f:
            write_component_file(f, snip)

    # Generate pages/*.py
    for snip in template.page_snippets:
        target_path = pages_dir / snip.name

        with target_path.open("w", encoding="utf-8") as f:
            write_component_file(f, snip)

    # Generate /*.py
    for snip in template.other_python_files:
        source_string = snip.stripped_code()
        target_path = main_module_dir / snip.name

        with target_path.open("w", encoding="utf-8") as f:
            f.write(source_string)

    # Find the main page
    homepage_snippet = template.homepage_snippet

    # Generate /project/__init__.py
    with open(main_module_dir / "__init__.py", "w", encoding="utf-8") as fil:
        generate_root_init(
            out=fil,
            raw_name=raw_name,
            project_type=type,
            components=template.component_snippets,
            pages=template.page_snippets,
            homepage_snippet=homepage_snippet,
            root_init_snippet=template.root_init_snippet,
            on_app_start=template.on_app_start,
            default_attachments=template.default_attachments,
        )

    # Generate /project/components/__init__.py
    with open(
        main_module_dir / "components" / "__init__.py", "w", encoding="utf-8"
    ) as f:
        write_init_file(f, template.component_snippets)

    # Generate /project/pages/__init__.py
    with open(
        main_module_dir / "pages" / "__init__.py", "w", encoding="utf-8"
    ) as f:
        write_init_file(f, template.page_snippets)

    # Generate a file specifying all dependencies, if there are any
    generate_dependencies_file(project_dir, template.dependencies)

    # Generate README.md
    with open(project_dir / "README.md", "w", encoding="utf-8") as f:
        generate_readme(f, raw_name, template)

    # Applications require a `__main__.py` as well
    if type == "app":
        with open(main_module_dir / "__main__.py", "w", encoding="utf-8") as f:
            f.write(
                f"""
# Make sure the project is in the Python path
import sys
from pathlib import Path
sys.path.insert(0, str(Path(__file__).parent.parent))


# Import the main module
import {module_name}

# Run the app
{module_name}.app.run_in_window()
"""
            )

    # Report success
    #
    # TODO: Other tools like poetry!? Add a command to activate the venv?
    print()
    success(f"The project has been created!")
    success(f"You can find it at `{project_dir.resolve()}`")
    print()
    print(f"To see your new project in action, run the following commands:")
    print()
    print(f"[dim]>[/] cd {revel.shell_escape(project_dir.resolve())}")

    if template.dependencies:
        print(
            f"[dim]>[/] python -m pip install -r requirements.txt  [bold]# Don't forget to install dependencies![/]"
        )

    print(f"[dim]>[/] rio run")



# <additional-imports>
import functools

import rio

from .. import data_models

# </additional-imports>


# <component>
class ItemList(rio.Component):
    """
    A component for displaying a list of menu items.

    Returns a list of menu items in a ListView component. Each item in the list
    contains the name, description.
    ## Attributes

    `menu_items`: The list of menu items to be displayed.

    `on_add_new_item_event`: An event handler for adding a new item.

    `on_delete_item_event`: An event handler for deleting an item.

    `on_select_item_event`: An event handler for selecting an item.
    """

    menu_items: list[data_models.MenuItem]
    on_add_new_item_event: rio.EventHandler[[]] = None
    on_delete_item_event: rio.EventHandler[int] = None
    on_select_item_event: rio.EventHandler[data_models.MenuItem] = None

    async def on_press_delete_item_event(self, idx: int) -> None:
        """
        Asynchronously triggers the 'delete item' when the delete button is
        pressed. The event handler is passed the index of the item to be
        deleted.

        ## Parameters

        `idx`: The index of the item to be deleted.
        """
        await self.call_event_handler(self.on_delete_item_event, idx)

        # Update the list
        await self.force_refresh()

    async def on_press_select_item_event(
        self, item: data_models.MenuItem
    ) -> None:
        """
        Asynchronously triggers the 'select item' when an item is selected. The
        event handler is passed the selected item.

        ## Parameters

        `item`: The selected item.
        """
        await self.call_event_handler(self.on_select_item_event, item)

    def build(self) -> rio.Component:
        """
        Builds the component by returning a ListView component containing the
        menu items.

        See the approx. layout below:

        ```
        ╔══════════════════════ ListView ══════════════════════╗
        ║ ┏━━━━━━━━━━━━━━━━━ SimpleListItem ━━━━━━━━━━━━━━━━━┓ ║
        ║ ┃ Add new                                          ┃ ║
        ║ ┗━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━┛ ║
        ║ ┏━━━━━━━━━━━━━━━━━ SimpleListItem ━━━━━━━━━━━━━━━━━┓ ║
        ║ ┃ Item 1                               ┏━Button━┓  ┃ ║
        ║ ┃                                      ┗━━━━━━━━┛  ┃ ║
        ║ ┗━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━┛ ║
        ║ ┏━━━━━━━━━━━━━━━━━ SimpleListItem ━━━━━━━━━━━━━━━━━┓ ║
        ║ ┃ Item 2                               ┏━Button━┓  ┃ ║
        ║ ┃                                      ┗━━━━━━━━┛  ┃ ║
        ║ ┗━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━┛ ║
        ║ ...                                                  ║
        ║ ┏━━━━━━━━━━━━━━━━━ SimpleListItem ━━━━━━━━━━━━━━━━━┓ ║
        ║ ┃ Item n                               ┏━Button━┓  ┃ ║
        ║ ┃                                      ┗━━━━━━━━┛  ┃ ║
        ║ ┗━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━┛ ║
        ╚══════════════════════════════════════════════════════╝
        ```
        """

        # Store all children in an intermediate list
        list_items = []

        list_items.append(
            rio.SimpleListItem(
                text="Add new",
                secondary_text="Description",
                key="add_new",
                left_child=rio.Icon("material/add"),
                on_press=self.on_add_new_item_event,
            )
        )

        for i, item in enumerate(self.menu_items):
            list_items.append(
                rio.SimpleListItem(
                    text=item.name,
                    secondary_text=item.description,
                    right_child=rio.Button(
                        rio.Icon("material/delete"),
                        color=self.session.theme.danger_color,
                        # Note the use of functools.partial to pass the
                        # index to the event handler.
                        on_press=functools.partial(
                            self.on_press_delete_item_event, i
                        ),
                    ),
                    # Use the name as the key to ensure that the list item
                    # is unique.
                    key=item.name,
                    # Note the use of functools.partial to pass the
                    # item to the event handler.
                    on_press=functools.partial(
                        self.on_press_select_item_event, item
                    ),
                )
            )

        # Then unpack the list to pass the children to the ListView
        return rio.ListView(
            *list_items,
            align_y=0,
        )


# </component>



from typing import *  # type:ignore

# <additional-imports>
import pandas as pd
import plotly.express as px

import rio

from .. import data_models

# </additional-imports>


# <component>
class CryptoChart(rio.Component):
    """
    The CryptoChart class is a component of a dashboard application, designed to
    handle and display cryptocurrency-related data.


    ## Attributes

    `data`: Historical data of the fetched crypto coins.

    `coin`: Name of the selected coin.

    `logo_url`: Representing the URL of the cryptocurrency's logo.

    `color`: Graph color of the selected coin.
    """

    data: pd.DataFrame
    coin: str
    logo_url: str = data_models.MY_COINS["bitcoin"][3]
    color: str = data_models.MY_COINS["bitcoin"][2]

    def on_change_coin(self, ev: rio.DropdownChangeEvent) -> None:
        """
        Updates the coin, color and logo_url attributes based on
        the selected coin.

        ## Parameters

        `ev`: The event object containing the selected coin value.
        """
        self.coin = ev.value
        self.color = data_models.MY_COINS[self.coin][2]
        self.logo_url = data_models.MY_COINS[self.coin][3]

    def build(self) -> rio.Component:
        """
        Creates a line plot of the last 50 data points of the selected coin.

        See the approx. layout below:
        ```
        ╔═════════════════════════════ CARD ══════════════════════════════╗
        ║ ┏━━━━━━━━━━━━━━━━━━━━━━━━━━ Column ━━━━━━━━━━━━━━━━━━━━━━━━━━━┓ ║
        ║ ┃ ┏━━━━━━━━━━━━━━━━━━━━━━━━━ Row ━━━━━━━━━━━━━━━━━━━━━━━━━━━┓ ┃ ║
        ║ ┃ ┃ ┏━━━━ Image ━━━━┓  ┏━━━━ Text ━━━━┓  ┏━━━ Dropdown ━━━┓ ┃ ┃ ║
        ║ ┃ ┃ ┃ url           ┃  ┃ str          ┃  ┃ dict           ┃ ┃ ┃ ║
        ║ ┃ ┃ ┗━━━━━━━━━━━━━━━┛  ┗━━━━━━━━━━━━━━┛  ┗━━━━━━━━━━━━━━━━┛ ┃ ┃ ║
        ║ ┃ ┗━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━┛ ┃ ║
        ║ ┃                                                             ┃ ║
        ║ ┃ ┏━━━━━━━━━━━━━━━━━━━━━━━━━ Plot ━━━━━━━━━━━━━━━━━━━━━━━━━━┓ ┃ ║
        ║ ┃ ┃                                                         ┃ ┃ ║
        ║ ┃ ┃ Plotly Express Figure                                   ┃ ┃ ║
        ║ ┃ ┃                                                         ┃ ┃ ║
        ║ ┃ ┃                                                         ┃ ┃ ║
        ║ ┃ ┃                                                         ┃ ┃ ║
        ║ ┃ ┗━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━┛ ┃ ║
        ║ ┗━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━┛ ║
        ╚═════════════════════════════════════════════════════════════════╝
        ```
        """

        fig = px.line(
            self.data[self.coin].iloc[-50:],
            color_discrete_sequence=[self.color],
            height=200,
            width=200,
        )

        # Set x-axis labels to horizontal and remove labels
        fig.update_xaxes(tickangle=0, title_text="")
        fig.update_yaxes(title_text="")

        # strip down the rest of the plot
        fig.update_layout(
            showlegend=False,
            margin=dict(t=10, l=10, b=10, r=10),
        )

        return rio.Card(
            rio.Column(
                rio.Row(
                    rio.Image(
                        rio.URL(self.logo_url),
                        height=2,
                        width=2,
                    ),
                    rio.Text(
                        self.coin.capitalize(),
                        style=rio.TextStyle(font_size=1.2, font_weight="bold"),
                    ),
                    rio.Dropdown(
                        options={
                            value[1]: key
                            for key, value in data_models.MY_COINS.items()
                        },
                        on_change=self.on_change_coin,
                    ),
                    spacing=1,
                    align_x=0.1,
                ),
                rio.Plot(
                    figure=fig,
                    corner_radius=0,
                    height=20,
                    width=10,
                    background=self.session.theme.neutral_color,
                    # margin=1,
                ),
                spacing=1,
                align_y=0.5,
                margin=2,
            ),
        )


# </component>



from __future__ import annotations

from typing import *  # type: ignore

import rio


# <component>
class NewsArticle(rio.Component):
    """
    Displays a news article with some visual separation from the background.
    """

    markdown: str

    def build(self) -> rio.Component:
        return rio.Card(
            rio.Markdown(
                self.markdown,
                margin=2,
            )
        )


# </component>



from __future__ import annotations

from dataclasses import KW_ONLY, dataclass
from typing import Any, final

from uniserde import JsonDoc

import rio.docs

from .fundamental_component import KeyboardFocusableFundamentalComponent

__all__ = [
    "TextInput",
    "TextInputChangeEvent",
    "TextInputConfirmEvent",
]


@final
@rio.docs.mark_constructor_as_private
@dataclass
class TextInputChangeEvent:
    """
    Holds information regarding a text input change event.

    This is a simple dataclass that stores useful information for when the user
    changes the text in a `TextInput`. You'll typically receive this as argument
    in `on_change` events.

    ## Attributes

    `text`: The new `text` of the `TextInput`.
    """

    text: str


@final
@rio.docs.mark_constructor_as_private
@dataclass
class TextInputConfirmEvent:
    """
    Holds information regarding a text input confirm event.

    This is a simple dataclass that stores useful information for when the user
    confirms the text in a `TextInput`. You'll typically receive this as
    argument in `on_confirm` events.

    ## Attributes

    `text`: The new `text` of the `TextInput`.
    """

    text: str


@final
class TextInput(KeyboardFocusableFundamentalComponent):
    """
    A user-editable text field.

    `TextInput` allows the user to enter a short text. The text can either be
    shown in plain text, or hidden when used for passwords or other sensitive
    information.


    ## Attributes

    `text`: The text currently entered by the user.

    `label`: A short text to display next to the text input.

    `prefix_text`: A short text to display before the text input. Useful for
            displaying currency symbols or other prefixed units.

    `suffix_text`: A short text to display after the text input. Useful for
            displaying units, parts of e-mail addresses, and similar.

    `is_secret`: Whether the text should be hidden. Use this to hide sensitive
            information such as passwords.

    `is_sensitive`: Whether the text input should respond to user input.

    `is_valid`: Visually displays to the user whether the current text is
            valid. You can use this to signal to the user that their input needs
            to be changed.

    `on_change`: Triggered when the user changes the text.

    `on_confirm`: Triggered when the user explicitly confirms their input,
            such as by pressing the "Enter" key. You can use this to trigger
            followup actions, such as logging in or submitting a form.


    ## Examples

    A simple `TextInput` with a default value of "John Doe" and a label. (Note:
    The value will not be updated if the user changes the value in the input
    field.)

    ```python
    rio.TextInput(text="John Doe", label="Name")
    ```

    You can easily bind state variables to track changes. If you want to make
    your TextInput more responsive, you can easily achieve this by adding a
    lambda function call to e.g. `on_change`:

    ```python
    class MyComponent(rio.Component):
        text: str = ""

        def build(self) -> rio.Component:
            return rio.TextInput(
                text=self.bind().text,  # attribute binding
                label="Name",
                on_change=lambda event: print(f"Name: {event.text}"),
            )
    ```

    You can also use a method for updating the input text and do whatever you
    want. Note that methods are handy if you want to do more than just updating
    the input text. For example run async code or update other components based
    on the input text:

    ```python
    class MyComponent(rio.Component):
        text: str = ""

        def on_change_update_text(self, event: rio.TextInputChangeEvent):
            self.text = event.text
            print(f"Name: {self.text}")

        def build(self) -> rio.Component:
            return rio.TextInput(
                text=self.text,
                label="Name",
                on_change=self.on_change_update_text,
            )
    ```
    """

    text: str = ""
    _: KW_ONLY
    label: str = ""
    prefix_text: str = ""
    suffix_text: str = ""
    is_secret: bool = False
    is_sensitive: bool = True
    is_valid: bool = True
    on_change: rio.EventHandler[TextInputChangeEvent] = None
    on_confirm: rio.EventHandler[TextInputConfirmEvent] = None

    def _validate_delta_state_from_frontend(self, delta_state: JsonDoc) -> None:
        if not set(delta_state) <= {"text"}:
            raise AssertionError(
                f"Frontend tried to change `{type(self).__name__}` state: {delta_state}"
            )

        if "text" in delta_state and not self.is_sensitive:
            raise AssertionError(
                f"Frontend tried to set `TextInput.text` even though `is_sensitive` is `False`"
            )

    async def _call_event_handlers_for_delta_state(
        self, delta_state: JsonDoc
    ) -> None:
        # Trigger on_change event
        try:
            new_value = delta_state["text"]
        except KeyError:
            pass
        else:
            assert isinstance(new_value, str), new_value
            await self.call_event_handler(
                self.on_change,
                TextInputChangeEvent(new_value),
            )

        self._apply_delta_state_from_frontend(delta_state)

    async def _on_message(self, msg: Any) -> None:
        # Listen for messages indicating the user has confirmed their input
        #
        # In addition to notifying the backend, these also include the input's
        # current value. This ensures any event handlers actually use the up-to
        # date value.
        assert isinstance(msg, dict), msg

        self._apply_delta_state_from_frontend({"text": msg["text"]})

        await self.call_event_handler(
            self.on_change,
            TextInputChangeEvent(self.text),
        )

        await self.call_event_handler(
            self.on_confirm,
            TextInputConfirmEvent(self.text),
        )

        # Refresh the session
        await self.session._refresh()


TextInput._unique_id = "TextInput-builtin"



from __future__ import annotations

import functools
from typing import *  # type: ignore

import rio

from .. import components as comps


# <code>
class TicTacToePage(rio.Component):
    # The contents of all fields. Each field can contain an X, an O, or be
    # empty. The initial state is an empty board.
    fields: list[Literal["X", "O", ""]] = [""] * 9

    # <new-attributes>
    # The player who is currently on turn
    player: Literal["X", "O"] = "X"
    # </new-attributes>

    # <on-press>
    async def on_press(self, index: int) -> None:
        """
        This function reacts to presses on the fields, and updates the game
        state accordingly.
        """

        # Set the field on that index
        self.fields[index] = self.player

        # Next player
        self.player = "X" if self.player == "O" else "O"

    # </on-press>

    # <on-reset>
    def on_reset(self) -> None:
        """
        Reset the game to its initial state.
        """
        self.fields = [""] * 9
        self.player = "X"

    # </on-reset>

    # <build>
    def build(self) -> rio.Component:
        # Spawn components for the fields
        field_components: list[rio.Component] = []

        for index, field in enumerate(self.fields):
            field_components.append(
                comps.Field(
                    value=field,
                    on_press=functools.partial(self.on_press, index),
                )
            )

        # Come up with a status message
        message = f"{self.player}'s turn"

        # Arrange all components in a grid
        return rio.Column(
            rio.Text(message, style="heading1"),
            rio.Grid(
                field_components[0:3],
                field_components[3:6],
                field_components[6:9],
                row_spacing=1,
                column_spacing=1,
                align_x=0.5,
            ),
            rio.Button(
                "Reset",
                icon="material/refresh",
                style="plain",
                on_press=self.on_reset,
            ),
            spacing=2,
            margin=2,
            align_x=0.5,
            align_y=0.0,
        )

    # </build>


# </code>



import rio.testing


async def test_active_page_url():
    url = "/foo/bar"

    async with rio.testing.TestClient(active_url=url) as test_client:
        assert test_client.session.active_page_url.path == url


async def test_crashed_build_functions_are_tracked():
    def build() -> rio.Component:
        return 3  # type: ignore

    async with rio.testing.TestClient(build) as test_client:
        assert len(test_client.crashed_build_functions) == 1


async def test_rebuild_resets_crashed_build_functions():
    class CrashingComponent(rio.Component):
        fail: bool = True

        def build(self) -> rio.Component:
            if self.fail:
                raise RuntimeError
            else:
                return rio.Text("hi")

    async with rio.testing.TestClient(CrashingComponent) as test_client:
        assert len(test_client.crashed_build_functions) == 1

        crashing_component = test_client.get_component(CrashingComponent)
        crashing_component.fail = False

        await test_client.refresh()

        assert not test_client.crashed_build_functions



from __future__ import annotations

from typing import *  # type: ignore

from uniserde import JsonDoc

import rio

from .fundamental_component import FundamentalComponent

__all__ = [
    "ThemeContextSwitcher",
]


@final
class ThemeContextSwitcher(FundamentalComponent):
    """
    A container which can switch between different components.

    A `ThemeContextSwitcher` is a container which can switch between different
    components. It is commonly used to switch between different themes. The
    `content` attribute can be used to change the currently displayed component.

    You can find more details on theming on the [theming how-to
    page](https://rio.dev/docs/howto/theming).


    ## Attributes

    `content`: The currently displayed component.

    `color`: The color of the switcher bar.


    ## Examples

    A minimal example of a `ThemeContextSwitcher`:

    ```python
    rio.ThemeContextSwitcher(
        content=rio.Button("Button"),
        color="secondary",
    )
    ```
    """

    content: rio.Component
    color: rio.ColorSet

    def _custom_serialize(self) -> JsonDoc:
        return {
            "color": self.session.theme._serialize_colorset(self.color),
        }


ThemeContextSwitcher._unique_id = "ThemeContextSwitcher-builtin"
